/* -*- C -*- */
#module CLX "V4.1"
/*------------------------------------------------------------*/
/*
 * CLX  "Command Language eXtractor"
 *
 * by Lennart Boerjeson, KTH/LNE, S-100 44 Stockholm, Sweden
 *
 * hacked by Richard Levitte, to remove the need for to almost identical
 * source files. <levitte@e.kth.se>
 *
 * CLX can really be compiled in two modes, depending on if the macro
 * DO_CLXX is defined or not.
 *
 * When compiled without DO_CLXX defined, CLX "disassembles" the currently
 * active CLI-table and prints a CLD-definition for any verb in the table.
 * When compiled with DO_CLXX defined, CLX "disassembles" the CLI-table
 * found in a given image file.
 *
 * Define CLX as a foreign command, e.g. CLX :== $disk:<dir>CLX.EXE
 * and call CLX by
 *
 * CLX [verb[ verb...]]		(for the version compiled without DO_CLXX)
 * CLXX file [verb[ verb...]]	(for the version compiled with DO_CLXX)
 */

#include <stdio.h>		/* Define standare io */
#include <ctype.h>		/* Define character functions */

#include "clitabdef.h"		/* dcl structures */

#ifdef DO_CLXX
#define CLITABLE clitable
#else  /* ! DO_CLXX */
#define CLITABLE ctl$ag_clitable
#endif /* DO_CLXX */

#ifdef DEBUG
#define Debug(commands) do {commands} while(0)
#else
#define Debug(commands)
#endif

/* Block relative offsets */
#define BRO(bl,bro) ((char *)bl+bl->bro)
/* like BRO, except if bro nil, return nil */
#define BRO_(bl,bro) (bl->bro ? BRO(bl,bro) : 0)
/* Table relative offset */
#define TRO(tro) ((char *)CLITABLE+(tro))

/*------------------------------------------------------------*/

/*
 * Program internal data structures
 */

/* Lists */

struct clinked_list {
  struct clinked_list *flink, *blink;
  char *data;
};

struct lbl_list
{
  struct lbl_list *flink, *blink;
  char num, *lbl;
};

struct par_list
{
  struct par_list *flink, *blink;
  char *ent, *parent;
};

#define insert(head,obj) \
{ obj->flink = head; obj->blink = head->blink; \
  head->blink->flink = obj; head->blink = obj; }
#define remove(obj) \
{ obj->blink->flink = obj->flink; \
  obj->flink->blink = obj->blink; }

/* The names of all predefined CLD types */

char *name_of_type[] =
{
  "$user_defined",
  "$infile",
  "$outfile",
  "$number",
  "$privilege",
  "$datetime",
  "$protection",
  "$process",
  "$inlog",
  "$outlog",
  "$insym",
  "$outsym",
  "$node",
  "$device",
  "$directory",
  "$uic",
  "$rest_of_line",
  "$parenthesized_value",
  "$deltatime",
  "$quoted_string",
  "$file",
  "$expression",
  "$$test1",
  "$$test2",
  "$$test3",
  "$acl"
};

/*------------------------------------------------------------*/

/*
 * The root block of the current CLI. This block contains references to the
 * following two blocks.
 */

#ifdef DO_CLXX
/* globalref */ struct vector_block	*clitable;
#else
globalref struct vector_block		*ctl$ag_clitable;
#endif

/*
 * This table is a table of all verb names. These are names the CLI
 * recognizes and acts upon. A reference to the corresponding command_block
 * is stored in the next table.
 */

struct vector_block *verb_name_table;

/*
 * A table of TROs to the command_blocks.
 */

struct vector_block *command_block_table;

/*------------------------------------------------------------*/

/*
 * CLX work in two phases. During phase one CLX scans the verb_name_table,
 * searching for the given verb, when it and its command_block is found,
 * a recursive search after referenced type- and syntax-blocks is made.
 * The referencies to these types, syntaxes and verbs are collected in
 * a linked list.
 *
 * During phase two the blocks in this list is decoded and printed.
 */

struct clinked_list *block_list;

/*
 * insert_block
 *
 * Inserts a block address in the linked list of blocks if
 * it isn't already there.
 */
insert_block(bl)
     char *bl;
{
  struct clinked_list *p;

  p = block_list->flink;
  while (p!=block_list)
    {
      if (p->data==bl) return 0;
      p = p->flink;
    };
  p = malloc(sizeof(struct clinked_list));
  p->data = bl;
  insert(block_list,p);
  return 1;
}

reinsert_block(bl)
     char *bl;
{
  struct clinked_list *p;

  p = block_list->flink;
  while (p!=block_list)
    {
      if (p->data==bl)
	{
	  remove(p);
	  insert(block_list,p);
	  break;
	}
      p = p->flink;
    };
}

/*------------------------------------------------------------*/

/*
 * All entity blocks have a number and a label. To simplify the decoding
 * of the outputs list, all numbers and labels of all entities encountered
 * are collected in this list. This collection is made during printout and
 * only the latest numbers are relevant.
 */

struct lbl_list *label_list;


/*
 * insert_label
 *
 * Inserts a number and a label in the linked list of labels.
 */
insert_label(num,lbl)
     int num;
     char *lbl;
{
  struct lbl_list *p;

  p = label_list->flink;
  while (p!=label_list)
    {
      if (p->num==num) break;
      p = p->flink;
    };
  if (p==label_list)
    {
      p = malloc(sizeof(struct lbl_list));
      p->num = num;
      insert(label_list,p);
    }
  p->lbl = lbl;
  return 1;
}

/*------------------------------------------------------------*/

/*
 * In order to decode the paths specified in the disallow expression,
 * we need to know under what DEFINE statement each entity was specified.
 * This list list stores all parents of all entities.
 *
 * Collected during Phase I.
 */

struct par_list *parent_list;
char *current_parent;

/*
 * insert_parent
 *
 * Inserts a parent address in the linked list of entities if
 * it isn't already there.
 */
insert_parent(bl)
     char *bl;
{
  struct par_list *p;

  p = parent_list->flink;
  while (p!=parent_list)
    {
      if (p->ent==bl) return p->parent;
      p = p->flink;
    };
  p = malloc(sizeof(struct par_list));
  p->ent = bl;
  p->parent = current_parent;
  insert(parent_list,p);
  return 0;
}

/*------------------------------------------------------------*/

/*
 * Output buffer manipulation. The output must be broken in lines, and the
 * the breaks should not appear in the middle of words. All output is sent
 * to outbuf, and check_outbuf determines weather the buffer should be
 * printed, partially printed or not printed at all.
 * The indentation logic is tricky.
 */

#define screen_width 80
#define total_indent (indent_width+extra_indent)
#define line_width (screen_width-total_indent)
#define maxbuflen 512
char outbuf[maxbuflen], medbuf[maxbuflen], sndbuf[maxbuflen];
char tmpbuf[maxbuflen];
int indent_width,split_indent,extra_indent;
int norm_indent = 8;

#define bufprintf(a) {sprintf(medbuf,a); check_outbuf}
#define bufprintf2(a,b) {sprintf(medbuf,a,b); check_outbuf}
#define bufprintf3(a,b,c) {sprintf(medbuf,a,b,c); check_outbuf}
#define bufprintf4(a,b,c,d) {sprintf(medbuf,a,b,c,d); check_outbuf}
#define bufprintf5(a,b,c,d,e) {sprintf(medbuf,a,b,c,d,e); check_outbuf}
#define check_outbuf {check_buf();}

#define set_indent(ind,spl) {indent_width = ind; split_indent = spl;}

/*
 * bufsubhead begins a verb, type or syntax clause.
 */

bufsubhead(str)
     char *str;
{
  set_indent(norm_indent,strlen(str));
  bufprintf(str);
}

/*
 * Check_buf checks what to do with the outbuf.
 */

check_buf()
{
  int len,pl;

  strcat(outbuf,medbuf);
  medbuf[0] = '\0';
  len = strlen(outbuf);
  if (len>line_width)
    split_buf(line_width);
  else
    {
      pl = len;
      len = strcspn(outbuf,"\n");
      if (len!=pl)
	{
	  send_buf(len+1);
	  extra_indent = 0;
	}
    }
}

send_buf(len)
     int len;
{
  int i;

  sndbuf[len] = '\0';
  strncpy(sndbuf,outbuf,len);
  for(i=len;outbuf[i]==' ';i++);
  strcpy(outbuf,outbuf+i);
  if (sndbuf[len-1] == '\n')
    sndbuf[len-1]='\0';
  printf("%*s%s\n",total_indent,"",sndbuf);
  extra_indent = split_indent;
  check_buf();
}

split_buf(at)
     int at;
{
  int len,i,sp,cm;

  len=at;

  sp = find_last(outbuf,at,' ');
  if (sp==at)
    {
      cm = find_last(outbuf,at,',');
      if (cm!=at)
	len=cm+1;
    }
  else
    len = sp;

  send_buf(len);
}

find_last(str,from,ch)
     char *str,ch;
{
  int i;

  i = from-1;
  while(--i>=0)
    {
      if (str[i]==ch)
	break;
    }
  if (i<0)
    i = from;
  return i;
}

/*------------------------------------------------------------*/
/*
 * Main, command interpreter
 */

main(argc, argv)
     int argc;
     char *argv[];
{
  int i;
#ifdef DO_CLXX
  char fname[maxbuflen];
  extern int get_clitable ()
#endif


  Debug(for (i = 1; i<argc; i++)
	printf((i < argc-1) ? "%s " : "%s\n", argv[i]););

  /*
   * Initialize the linked lists.
   */

  block_list = malloc(sizeof(struct clinked_list));
  block_list->flink = block_list;
  block_list->blink = block_list;

  label_list = malloc(sizeof(struct lbl_list));
  label_list->flink = label_list;
  label_list->blink = label_list;

  parent_list = malloc(sizeof(struct par_list));
  parent_list->flink = parent_list;
  parent_list->blink = parent_list;

#ifdef DO_CLXX
  /*
   * Construct string descriptor to first arg.
   */

  if (argc<=1)
    {
      printf("%s","Enter filename: ");
      scanf("%s",fname);
    }
  else
    {
      strcpy(fname,argv[1]);
      argc--;
      argv++;
    };
#endif

  /*
   * Initialize the pointers to the verb- and command-table.
   */

#ifdef DO_CLXX
  get_clitable(&clitable,fname);

  if (clitable==0)
    {
      bufprintf2("%%CLXX-E-NOCLI  No CLI table found in file %s\n",fname);
      return;
    }
#endif

  verb_name_table = TRO(CLITABLE->vec_l_verbtbl);
  command_block_table = TRO(CLITABLE->vec_l_comdptr);

  /*
   * Check the structure level. Supported level is 6.
   */

  if (CLITABLE->vec_r_flags.vec_b_strlvl!=vec_c_strlvl)
    {
      bufprintf2("%%CLX-E-UNSUP   Unsupported structure level (%d)\n",
		 CLITABLE->vec_r_flags.vec_b_strlvl);
      bufprintf2("%%CLI-I-CURLVL  Current structure level = %d\n",vec_c_strlvl);
      return;
    }

  Debug(print_header(CLITABLE););
  Debug(print_header(verb_name_table););
  Debug(print_header(command_block_table););

  /*
   * Collect verbs, syntaxes and types.
   *
   * If no argument, find all verbs.
   * Otherwise find arguments.
   */

  if (argc<=1) enter_command("");
  else for (i=1;i<argc;i++)
    {
      enter_command(argv[i]);
    }

  /*
   * Decode them!
   */

  print_commands();
}

/*------------------------------------------------------------*/
/*
 * enter_command
 *
 * Implements phase one.
 */

enter_command(cmd)
     char *cmd;
{
  int i,max_i;
  char (*comm)[][4];
  int (*comb)[];

  /*
   * Scan the verb table.
   */

  upcase(cmd);
  comm = &(verb_name_table->vec_l_verbtbl);
  comb = &(command_block_table->vec_l_verbtbl);
  max_i = command_block_table->vec_w_tro_count;

  for(i=0; i<max_i; i++)
    if ((strlen(cmd)==0)||(strncmp(cmd,(*comm)[i],4)==0))
      {
	Debug(bufprintf2("Cmd=%.4s, ",(*comm)[i]););
	Debug(print_header(TRO((*comb)[i])););

	/*
	 * Enter this block and search it for referencies to other blocks.
	 */

	enter_block(TRO((*comb)[i]));
      }
}

/*------------------------------------------------------------*/
/*
 * enter_block
 *
 * Enters a block and all referenced syntax- and type-blocks
 * in the the list of blocks to be printed.
 */
enter_block(bl)
     struct vector_block *bl;
{
  /*
   * Insert only command- and type-blocks.
   * Scan them only if successfully inserted.
   */

  switch(bl->vec_b_type)
    {
    case block_c_vector      : break;
    case block_c_command     : if (insert_block(bl))
      {
	current_parent = bl;
	scan_command(bl);
	reinsert_block(bl);	/* Move to end of list */
      }
      break;
    case block_c_type        : if (insert_block(bl))
      {
	current_parent = bl;
	scan_type(bl);
	reinsert_block(bl);	/* Move to end of list */
      }
      break;
    case block_c_entity      : if (!insert_parent(bl))
      {
	scan_entity(bl);
      }
      break;
    case block_c_expression  : break;
    case block_c_cdu_visited : break;
    };
}

scan_command(bl)
     struct command_block *bl;
{
  if (bl->cmd_b_type==block_c_command)
    {
      if (bl->cmd_r_flags.cmd_v_parms)
	enter_block(TRO(bl->cmd_l_parms));
      if (bl->cmd_r_flags.cmd_v_quals)
	enter_block(TRO(bl->cmd_l_quals));
    }
}

scan_type(bl)
     struct type_block *bl;
{
  if (bl->type_b_type==block_c_type)
    enter_block(TRO(bl->type_l_keywords));
}

scan_entity(bl)
     struct entity_block *bl;
{
  char *my_parent;

  if (bl->ent_b_type==block_c_entity)
    {
      my_parent = current_parent;
      if (bl->ent_l_syntax!=0)
	enter_block(TRO(bl->ent_l_syntax));
      current_parent = my_parent;
      if (bl->ent_l_user_type!=0)
	enter_block(TRO(bl->ent_l_user_type));
      current_parent = my_parent;
      if (bl->ent_l_next!=0)
	enter_block(TRO(bl->ent_l_next));
    }
}

/*------------------------------------------------------------*/
/*
 * print_commands
 *
 * Implements phase two.
 */

print_commands()
{
  struct clinked_list *p;

  p = block_list->flink;
  while (p!=block_list)
    {
      Debug(print_header(p->data););
      print_block(p->data);
      p = p->flink;
    }
}

/*------------------------------------------------------------*/
/*
 * print_block
 *
 * Prints the CLD equivalent of a block.
 */

print_block(bl)
     struct vector_block *bl;
{
  current_parent = bl;
  switch(bl->vec_b_type)
    {
    case block_c_vector      : bufprintf("! Can't print vector block\n");
      break;
    case block_c_command     : clx_command(bl);
      break;
    case block_c_type        : clx_type(bl);
      break;
    case block_c_entity      : bufprintf("! Can't print entity block\n");
      break;
    case block_c_expression  : bufprintf("! Can't print expression block\n");
      break;
    case block_c_cdu_visited : bufprintf("! Can't print cdu_visited block\n");
      break;
    };
  bufprintf("\n");
}

clx_command(bl)
     struct command_block *bl;
{
  int i,n;

  set_indent(0,norm_indent);
  switch(bl->cmd_b_subtype)
    {
    case cmd_c_verb   :
      bufprintf("define verb ");
      n = print_ascics(BRO(bl,cmd_w_name),1,1,"");
      bufprintf("\n");

      if (n>1)
	for (i=2; i<=n; i++)
	  {
	    bufsubhead("synonym ");
	    print_ascics(BRO(bl,cmd_w_name),i,i,"");
	    bufprintf("\n");
	  }
      break;
    case cmd_c_syntax :
      bufprintf("define syntax ");
      print_ascic(BRO(bl,cmd_w_name));
      bufprintf("\n");
      break;
    }

  switch(bl->cmd_b_handler)
    {
    case cmd_c_none :
      break;
    case cmd_c_cli :
      bufsubhead("cliroutine ");
      print_ascic(BRO(bl,cmd_w_image));
      bufprintf("\n");
      break;
    case cmd_c_user :
      bufsubhead("routine ");
      print_ascic(BRO(bl,cmd_w_image+4));
      bufprintf("\n");
      break;
    case cmd_c_image :
      bufsubhead("image ");
      print_ascic(BRO(bl,cmd_w_image));
      bufprintf("\n");
      break;
    case cmd_c_same :
      break;
    }

  if (bl->cmd_w_prefix!=0)
    {
      bufsubhead("prefix ");
      print_ascic(BRO(bl,cmd_w_prefix));
      bufprintf("\n");
    }

#define fl(flg) (bl->cmd_r_flags.flg)
#define iffls(f1,f2,f3,f4,f5) (fl(f1)+fl(f2)+fl(f3)+fl(f4)+fl(f5))
#define pfl(flg,str) if(fl(flg)) {if(i++!=0) bufprintf(","); bufprintf(str);}

  n=iffls(cmd_v_abbrev,cmd_v_nostat,cmd_v_foreign,cmd_v_immed,cmd_v_mcrparse);
  if(n!=0)
    {
      bufsubhead("cliflags(");
      i=0;
      pfl(cmd_v_abbrev,"abbreviate");
      pfl(cmd_v_nostat,"nostatus");
      pfl(cmd_v_foreign,"foreign");
      pfl(cmd_v_immed,"immediate");
      pfl(cmd_v_mcrparse,"mcrparse");
      bufprintf(")\n");
    }

#undef iffls
#undef fl
#undef pfl

  if (bl->cmd_r_flags.cmd_v_parms)
    if (bl->cmd_l_parms!=0)
      clx_entity(TRO(bl->cmd_l_parms));
    else
      bufsubhead("noparameters\n");

  if (bl->cmd_r_flags.cmd_v_quals)
    if (bl->cmd_l_quals!=0)
      clx_entity(TRO(bl->cmd_l_quals));
    else
      bufsubhead("noqualifiers\n");

  if (bl->cmd_r_flags.cmd_v_disallows)
    if (bl->cmd_l_disallow!=0)
      clx_disallow(TRO(bl->cmd_l_disallow));
    else
      bufsubhead("nodisallows\n");

  if (bl->cmd_w_outputs!=0)
    {
      bufsubhead("outputs   ");
      bufprintf("(");
      clx_outputs(BRO(bl,cmd_w_outputs));
      bufprintf(")\n");
    }
}

clx_outputs(cnt)
     char *cnt;
{
  char *i;
  int not_first_time;

  not_first_time=0;
  for (i=cnt+1; i<=cnt+(*cnt); i++)
    {
      if (not_first_time)
	bufprintf(",");
      clx_label_n(*i);
      not_first_time=1;
    }
}

clx_label_n(n)
     char n;
{
  struct lbl_list *p;
  char *lbl;

  lbl = 0;
  p = label_list->flink;
  while (p!=label_list)
    {
      if (p->num==n)
	{
	  lbl = p->lbl;
	  break;
	}
      p=p->flink;
    }
  if (lbl!=0)
    print_ascic(lbl);
  else
    bufprintf2("%d",n);
}

clx_type(bl)
     struct type_block *bl;
{
  set_indent(0,norm_indent);
  bufprintf("define type ");
  print_ascic(BRO(bl,type_w_name));
  bufprintf("\n");

  if (bl->type_w_prefix!=0)
    {
      bufsubhead("prefix ");
      print_ascic(BRO(bl,type_w_prefix));
      bufprintf("\n");
    }

  if (bl->type_l_keywords!=0)
    clx_entity(TRO(bl->type_l_keywords));
}

#define flags(fl) ((*((short *) &(bl->ent_r_flags)))&(fl))
#define bufprintfl(str) {bufprintf(delim); bufprintf(str); delim = ",";}

clx_entity(bl)
     struct entity_block *bl;
{
  int type_specified, default_specified;
  char *delim;
  int place;

  switch(bl->ent_b_subtype)
    {
    case ent_c_parameter :
      bufsubhead("parameter ");
      insert_label(-1*bl->ent_b_number,BRO_(bl,ent_w_label));
      break;
    case ent_c_qualifier :
      bufsubhead("qualifier ");
      insert_label(bl->ent_b_number,BRO_(bl,ent_w_label));
      break;
    case ent_c_keyword :
      bufsubhead("keyword ");
      break;
    }

  print_ascic(BRO(bl,ent_w_name));

  if (bl->ent_l_syntax)
    {
      bufprintf(", syntax=");
      print_cmd_name(TRO(bl->ent_l_syntax));
    }

  if (bl->ent_w_label)
    if (bl->ent_w_label!=bl->ent_w_name)
      {
	bufprintf(", label=");
	print_ascic(BRO(bl,ent_w_label));
      }

  if (bl->ent_w_prompt)
    {
      bufprintf(", prompt=\"");
      print_ascic(BRO(bl,ent_w_prompt));
      bufprintf("\"");
    }

  place = (flags(ent_m_verb)!=0)*2 + (flags(ent_m_parm)!=0);
  switch (place)
    {
    case 0 :
      break;
    case 1 :
      bufprintf(", placement=local");
      break;
    case 2 :
      bufprintf(", placement=global");
      break;
    case 3 :
      bufprintf(", placement=positional");
    }

  if flags(ent_m_val)
    {
      bufprintf(", value");
      type_specified = (bl->ent_l_user_type!=0) || (bl->ent_b_valtype!=0);
      default_specified =
	(bl->ent_w_defval ? print_ascics(BRO(bl,ent_w_defval),1,0,"") : 0);
      if (flags(ent_m_valreq|ent_m_list|ent_m_concat|ent_m_impcat)
	  || type_specified || default_specified)
	{
	  bufprintf("(");
	  delim = "";
	  if (!flags(ent_m_concat))
	    bufprintfl("noconcatenate");
	  if flags(ent_m_impcat)
		    bufprintfl("impcat");
	  if flags(ent_m_list)
		    bufprintfl("list");
	  if flags(ent_m_valreq)
		    bufprintfl("required");
	  if (type_specified)
	    {
	      bufprintfl("type=");
	      if (bl->ent_b_valtype==0)
		print_type_name(TRO(bl->ent_l_user_type));
	      else
		bufprintf(name_of_type[bl->ent_b_valtype]);
	    }
	  if (default_specified)
	    {
	      bufprintfl("default=");
	      if (default_specified>1)
		bufprintf("(");
	      print_ascics(BRO(bl,ent_w_defval),1,default_specified,"\"");
	      if (default_specified>1)
		bufprintf(")");
	    }
	  bufprintf(")");
	}
    }

  switch(bl->ent_b_subtype)
    {
    case ent_c_parameter :
      if (flags(ent_m_neg))
	{
	  bufprintf(", negatable");
	}
      break;
    case ent_c_qualifier :
      if (!flags(ent_m_neg))
	{
	  bufprintf(", nonnegatable");
	}
      break;
    case ent_c_keyword :
      if (flags(ent_m_neg))
	{
	  bufprintf(", negatable");
	}
      break;
    }

  if flags(ent_m_deftrue)
    {
      bufprintf(", default");
    }

  if flags(ent_m_batdef)
    {
      bufprintf(", batch");
    }

  if flags(ent_m_mcroptdelim|ent_m_mcrignore)
    {
      bufprintf4(", cliflags(%s%s%s)",
		 (flags(ent_m_mcroptdelim) ? "mcroptdelim" : ""),
		 (flags(ent_m_mcroptdelim)&&flags(ent_m_mcrignore) ? "," : ""),
		 (flags(ent_m_mcrignore) ? "mcrignore" : ""));
    }

  bufprintf("\n");

  if(bl->ent_l_next!=0)
    clx_entity(TRO(bl->ent_l_next));
}

#undef flags
#undef bufprintfl

clx_disallow(bl)
     struct expression_block *bl;
{
  int i;
  long (*operand)[];

  operand = &(bl->exp_l_operand_list);

  if (bl->exp_b_subtype!=exp_c_or)
    {
      bufsubhead("disallow  ");
      clx_expression(bl,3,0,0);
      bufprintf("\n");
    }
  else
    for (i=0; i<bl->exp_w_tro_count; i++)
      {
	bufsubhead("disallow  ");
	clx_expression(TRO((*operand)[i]),3,0,0);
	bufprintf("\n");
      }
}

struct op_rec {
  int prio;
  char *name;
  char *l, *m, *r;
} op_data[] =
  {
    {0,"","","",""},
    {0,"path","","",""},
    {1,"not","not ","",""},
    {1,"any2","any2(",",",")"},
    {2,"and",""," and ",""},
    {3,"or",""," or ",""},
    {3,"xor",""," xor ",""},
    {1,"neg","neg ","",""}
  };

clx_expression(bl,pri,dlm,ii)
     struct expression_block *bl;
     int pri;
     char **dlm;
{
  int i;
  struct op_rec d;
  long (*operand)[];
  char *my_parent;

  switch(bl->exp_b_type)
    {
    case block_c_command :
      if (ii==0)
	{
	  bufprintf("<");
	  print_cmd_name(bl);
	  bufprintf(">");
	  if (dlm!=0)
	    *dlm = "";
	}
      break;
    case block_c_entity :
      my_parent = insert_parent(bl);
      if (my_parent!=current_parent && ii==0)
	{
	  bufprintf("<")
	    print_def_name(my_parent);
	  bufprintf(">");
	}
      print_ent_label(bl);
      if (dlm!=0)
	*dlm = ".";
      break;
    case block_c_expression :
      d = op_data[bl->exp_b_subtype];
      operand = &(bl->exp_l_operand_list);
      if (d.prio>pri)
	bufprintf("(");
      bufprintf(d.l);
      for(i=0; i<bl->exp_w_tro_count; i++)
	{
	  if (i!=0)
	    bufprintf(d.m);
	  clx_expression(TRO((*operand)[i]),d.prio,&d.m,i);
	}
      bufprintf(d.r);
      if (d.prio>pri)
	bufprintf(")");
      break;
    }
}

/*------------------------------------------------------------*/
/*
 * Print the header block.
 * Intended for debug.
 */

print_header(tp)
     struct vector_block *tp;
{
  bufprintf5("! Address %8x, Size %x, type %d, subtype %d\n",
	     tp, tp->vec_w_size, tp->vec_b_type, tp->vec_b_subtype);
}

/*---------------------------------------------------@/
/*
 * Common subroutines.
 */

/*
 * Upcase
 *
 * Converts a string to uppercase.
 */

upcase(str)
     char *str;
{
  register int i,l;

  l = strlen(str);
  for (i=0;i<l;i++)
    {
      *str = _toupper(*str);
      str++;
    }
}

/*
 * print_ascic
 *
 * Prints an ASCIC string
 */

print_ascic(ascic)
     char *ascic;
{
  bufprintf3("%.*s",*ascic,ascic+1);
}

/*
 * print_ascics
 *
 * Prints a sequence of ASCIC strings. The sequence is itself an ASCIC string.
 *
 * The parameters first and last controls which strings to print.
 */

print_ascics(ascics,first,last,dlm)
     char *ascics, *dlm;
     int first,last;
{
  char *ascic;
  char not_first_time;
  int i;

  not_first_time = 0;
  i = 0;
  for (ascic=ascics+1; ascic<ascics+(*ascics); ascic=ascic+(*ascic)+1)
    {
      if (*ascic!=0)
	{
	  i++;
	  if ((i>=first)&&(i<=last))
	    {
	      if (not_first_time)
		bufprintf(",");
	      bufprintf(dlm);
	      print_ascic(ascic);
	      bufprintf(dlm);
	      not_first_time=1;
	    }
	}
    }
  return i;
}

print_ent_name(bl)
     struct entity_block *bl;
{
  print_ascic(BRO(bl,ent_w_name));
}

print_ent_label(bl)
     struct entity_block *bl;
{
  print_ascic(BRO(bl,ent_w_label));
}

print_type_name(bl)
     struct type_block *bl;
{
  print_ascic(BRO(bl,type_w_name));
}

print_cmd_name(bl)
     struct command_block *bl;
{
  switch(bl->cmd_b_subtype)
    {
    case cmd_c_verb :
      print_ascics(BRO(bl,cmd_w_name),1,1,"");
      break;
    case cmd_c_syntax :
      print_ascic(BRO(bl,cmd_w_name));
      break;
    }
}

print_def_name(bl)
     struct vector_block *bl;
{
  switch(bl->vec_b_type)
    {
    case block_c_vector      : break;
    case block_c_command     : print_cmd_name(bl);
      break;
    case block_c_type        : print_type_name(bl);
      break;
    case block_c_entity      : break;
    case block_c_expression  : break;
    case block_c_cdu_visited : break;
    }
}
